import math
import copy
import time
import numpy as np
import pandas as pd

import torch
from torch import nn
import torch.nn.functional as F
from torchsummary import summary
import torch.backends.cudnn as cudnn
import torch.optim as optim


class conv_block(nn.Module):

  def __init__(self, in_channels, out_channels,  **kwargs):
    super(conv_block, self).__init__()
    self.relu = nn.ReLU()
    self.bathcnorm = nn.BatchNorm1d(out_channels)
    self.conv = nn.Conv1d(in_channels, out_channels, **kwargs)


  def forward(self, x):
    return self.relu(self.bathcnorm(self.conv(x)))

class conv_block_2(nn.Module):

  def __init__(self, in_channels, out_channels,  **kwargs):
    super(conv_block_2, self).__init__()
    self.block = nn.Sequential(
        nn.Conv1d(in_channels=in_channels, out_channels=out_channels, **kwargs),
        nn.BatchNorm1d(out_channels)
    )


  def forward(self, x):
    return self.block(x)


class Classifier_RESNET(nn.Module):

    def __init__(self, input_shape, nb_classes):
        super(Classifier_RESNET, self).__init__()
        self.nb_classes = nb_classes
        n_feature_maps = 64
        
        # BLOCK 1
        self.conv_x = conv_block(in_channels=1,                out_channels=n_feature_maps, kernel_size=8,  stride=1, padding='same')
        self.conv_y = conv_block(in_channels=n_feature_maps,   out_channels=n_feature_maps, kernel_size=5,  stride=1, padding='same')
        self.conv_z = conv_block_2(in_channels=n_feature_maps, out_channels=n_feature_maps, kernel_size=3, stride=1, padding='same')
        # expand channels for the sum
        self.shortcut_y = conv_block_2(in_channels=1, out_channels=n_feature_maps, kernel_size=1, stride=1, padding='same')
        self.relu = nn.ReLU()
    
        # BLOCK 2
        self.conv_x_2 = conv_block(in_channels=n_feature_maps,   out_channels=n_feature_maps*2, kernel_size=8,  stride=1, padding='same')
        self.conv_y_2 = conv_block(in_channels=n_feature_maps*2,   out_channels=n_feature_maps*2, kernel_size=5,  stride=1, padding='same')
        self.conv_z_2 = conv_block_2(in_channels=n_feature_maps*2, out_channels=n_feature_maps*2, kernel_size=3, stride=1, padding='same')
        # expand channels for the sum
        self.shortcut_y_2 = conv_block_2(in_channels=n_feature_maps, out_channels=n_feature_maps*2, kernel_size=1, stride=1, padding='same')
        self.relu = nn.ReLU()

        # BLOCK 3
        self.conv_x_3 = conv_block(in_channels=n_feature_maps*2,   out_channels=n_feature_maps*2, kernel_size=8,  stride=1, padding='same')
        self.conv_y_3 = conv_block(in_channels=n_feature_maps*2,   out_channels=n_feature_maps*2, kernel_size=5,  stride=1, padding='same')
        self.conv_z_3 = conv_block_2(in_channels=n_feature_maps*2, out_channels=n_feature_maps*2, kernel_size=3, stride=1, padding='same')
        # expand channels for the sum
        self.shortcut_y_3 = nn.BatchNorm1d(n_feature_maps*2)
        self.relu = nn.ReLU()


        self.avgpool1 = nn.AdaptiveAvgPool1d(1)
        self.fc1 = nn.Linear(2*n_feature_maps, self.nb_classes)

    def forward(self, x):
        input_layer = x

        # BLOCK 1
        conv_x = self.conv_x(x)
        conv_y = self.conv_y(conv_x)
        conv_z = self.conv_z(conv_y)
        shortcut_y = self.shortcut_y(input_layer)

        output_block_1 = shortcut_y + conv_z 
        output_block_1 = self.relu(output_block_1)
        # BLOCK 2
        conv_x_2 = self.conv_x_2(output_block_1)
        conv_y_2 = self.conv_y_2(conv_x_2)
        conv_z_2 = self.conv_z_2(conv_y_2)
        shortcut_y_2 = self.shortcut_y_2(output_block_1)
        output_block_2 = shortcut_y_2 + conv_z_2
        output_block_2 = self.relu(output_block_2)

        # BLOCK 3
        conv_x_3 = self.conv_x_3(output_block_2)
        conv_y_3 = self.conv_y_3(conv_x_3)
        conv_z_3 = self.conv_z_3(conv_y_3)
        shortcut_y_3 = self.shortcut_y_3(output_block_2)
        output_block_3 = shortcut_y_3 + conv_z_3
        output_block_3 = self.relu(output_block_3)

        x = self.avgpool1(output_block_3)
        x = x.reshape(x.shape[0], -1)
        x = self.fc1(x)

        return x




def fit(trainloader, valloader, input_shape, nb_classes):
    print('Hi from RESNET!')
    model = Classifier_RESNET(input_shape, nb_classes)
    print(model)
    use_cuda = torch.cuda.is_available()                
    if use_cuda:
        torch.cuda.set_device(0)
        model.cuda()
        cudnn.benchmark = True
    summary(model, ( input_shape[-2], input_shape[-1]))

    # Training
    def train_alone_model(net, epoch):

        print('\Teaining epoch: %d' % epoch)
        net.train()

        train_loss = 0
        correct = 0
        total = 0
        
        for batch_idx, (inputs, targets) in enumerate(trainloader):
            if use_cuda:
                inputs, targets = inputs.cuda(), targets.cuda()
            optimizer.zero_grad()

            outputs = net(inputs.float())
            loss = criterion_CE(outputs, targets)
            loss.backward() 
            optimizer.step()  

            train_loss += loss.item()

            _, predicted = torch.max(outputs.data, 1)
            total += targets.size(0)
            correct += predicted.eq(targets.data).cpu().sum().float().item()

            b_idx = batch_idx

        print('Training Loss: %.3f | Training Acc: %.3f%% (%d/%d)' % (train_loss / (b_idx + 1), 100. * correct / total, correct, total))

        return train_loss / (b_idx + 1)

    def test(net):
        net.eval()
        test_loss = 0
        correct = 0
        total = 0
        for batch_idx, (inputs, targets) in enumerate(valloader):
            if use_cuda:
                inputs, targets = inputs.cuda(), targets.cuda()
            outputs = net(inputs.float())
            loss = criterion_CE(outputs, targets)
            test_loss += loss.item()
            _, predicted = torch.max(outputs.data, 1)
            total += targets.size(0)
            correct += predicted.eq(targets.data).cpu().sum().float().item()
            b_idx = batch_idx

        print('Test Loss: %.3f | Test Acc: %.3f%% (%d/%d)' % (test_loss / (b_idx + 1), 100. * correct / total, correct, total))
        return test_loss / (b_idx + 1), correct / total

    final_loss = []
    learning_rates = []
    
    epochs = 20
    criterion_CE = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001,)
    scheduler = optim.lr_scheduler.OneCycleLR(optimizer, 1e-3, epochs=20, steps_per_epoch=len(trainloader))

    use_cuda = torch.cuda.is_available()        
    best_model_wts = copy.deepcopy(model.state_dict())
    min_train_loss = np.inf

    start_time = time.time()    
    for epoch in range(epochs):
        train_loss = train_alone_model(model, epoch)
        test(model)
        if min_train_loss  > train_loss:
            min_train_loss = train_loss
            best_model_wts = copy.deepcopy(model.state_dict())
        scheduler.step()
    
        final_loss.append(train_loss)
        learning_rates.append(optimizer.param_groups[0]['lr'])
    
    output_directory = '/'
    torch.save(best_model_wts, output_directory +  'best_teacher_model.pt')


    # Save Logs
    duration = time.time() - start_time
    best_teacher = Classifier_RESNET(input_shape, nb_classes)
    best_teacher.load_state_dict(best_model_wts)
    best_teacher.cuda()
    
    print('Best Model Accuracy in below ')
    start_test_time = time.time()
    test(best_teacher)
    test_duration = time.time() - start_test_time

    print(test(best_teacher))

    df = pd.DataFrame(list(zip(final_loss, learning_rates)), columns =['loss', 'learning_rate'])
    index_best_model = df['loss'].idxmin()
    row_best_model = df.loc[index_best_model]
    df_best_model = pd.DataFrame(list(zip([row_best_model['loss']], [index_best_model+1])), columns =['best_model_train_loss', 'best_model_nb_epoch'])

    df.to_csv(output_directory + 'history.csv', index=False)
    df_best_model.to_csv(output_directory + 'df_best_model.csv', index=False)

    loss_, acc_ = test(best_teacher)
    df_metrics = pd.DataFrame(list(zip([min_train_loss], [acc_], [duration], [test_duration])), columns =['Loss', 'Accuracy', 'Duration', 'Test Duration'])
    df_metrics.to_csv(output_directory + 'df_metrics.csv', index=False)




import pandas as pd
import numpy as np
import os
from importlib import resources
import sklearn
from sklearn.preprocessing import LabelEncoder
import torch
from torch.utils.data import DataLoader, TensorDataset


def read_all_datasets(DATASET_NAMES_2018) -> pd.DataFrame:
    datasets_dict = {}

    for dataset_name in DATASET_NAMES_2018:
        with resources.path("romaniya_menim.data", dataset_name + "_TRAIN.tsv") as df:
            df_train =  pd.read_csv(df, sep='\t', header=None)
        with resources.path("romaniya_menim.data", dataset_name + "_TEST.tsv") as df:
            df_test =  pd.read_csv(df, sep='\t', header=None)

        y_train = df_train.values[:, 0]
        y_test = df_test.values[:, 0]
        
        x_train = df_train.drop(columns=[0])
        x_test = df_test.drop(columns=[0])
        
        x_train.columns = range(x_train.shape[1])
        x_test.columns = range(x_test.shape[1])
        
        x_train = x_train.values
        x_test = x_test.values
        
        # znorm
        std_ = x_train.std(axis=1, keepdims=True)
        std_[std_ == 0] = 1.0
        x_train = (x_train - x_train.mean(axis=1, keepdims=True)) / std_
        
        std_ = x_test.std(axis=1, keepdims=True)
        std_[std_ == 0] = 1.0
        x_test = (x_test - x_test.mean(axis=1, keepdims=True)) / std_
        
        datasets_dict[dataset_name] = (x_train.copy(), y_train.copy(), x_test.copy(), y_test.copy())


    return datasets_dict

def prepare_data(datasets_dict, dataset_name, classifier_name):
    x_train = datasets_dict[dataset_name][0]
    y_train = datasets_dict[dataset_name][1]
    x_test = datasets_dict[dataset_name][2]
    y_test = datasets_dict[dataset_name][3]

    nb_classes = len(np.unique(np.concatenate((y_train, y_test), axis=0)))

    # make the min to zero of labels
    y_train, y_test = transform_labels(y_train, y_test)

    # save orignal y because later we will use binary
    y_true = y_test.astype(np.int64)
    y_true_train = y_train.astype(np.int64)
    # transform the labels from integers to one hot vectors
    enc = sklearn.preprocessing.OneHotEncoder()

    if len(x_train.shape) == 2:  # if univariate
        # add a dimension to make it multivariate with one dimension
        x_train = x_train.reshape((x_train.shape[0], x_train.shape[1], 1))
        x_test = x_test.reshape((x_test.shape[0], x_test.shape[1], 1))

    new_x_train = np.transpose(x_train, (0, 2, 1))
    new_x_test = np.transpose(x_test, (0, 2, 1))
    
    x_train, y_train= torch.from_numpy(new_x_train), torch.from_numpy(y_train)
    x_test, y_test= torch.from_numpy(new_x_test), torch.from_numpy(y_test)
    input_shape = x_train.shape[1:]

    print('CLassifier name: ', classifier_name)
    
    if classifier_name == 'fcn':
        mini_batch_size = int(min(x_train.shape[0]/10, 16))
    elif classifier_name == 'cnn':
        mini_batch_size = 16
    elif classifier_name == 'mlp':
        mini_batch_size = int(min(x_train.shape[0]/10, 16))
    elif classifier_name == 'resnet':
        mini_batch_size = int(min(x_train.shape[0]/10, 16))

    elif classifier_name == 'inception':
        mini_batch_size = 64


    trainloader = DataLoader(
        TensorDataset(x_train, y_train),
        batch_size=mini_batch_size,
        shuffle=True,
    )
    valloader = DataLoader(
        TensorDataset(x_test, y_test),
        batch_size=mini_batch_size,
        shuffle=False
    )

    return trainloader, valloader, input_shape, nb_classes


def transform_labels(y_train, y_test):
    """
    Transform label to min equal zero and continuous
    For example if we have [1,3,4] --->  [0,1,2]
    """
    # no validation split
    # init the encoder
    encoder = LabelEncoder()
    # concat train and test to fit
    y_train_test = np.concatenate((y_train, y_test), axis=0)
    # fit the encoder
    encoder.fit(y_train_test)
    # transform to min zero and continuous labels
    new_y_train_test = encoder.transform(y_train_test)
    # resplit the train and test
    new_y_train = new_y_train_test[0:len(y_train)]
    new_y_test = new_y_train_test[len(y_train):]
    return new_y_train, new_y_test



import math
import copy
import time
import numpy as np
import pandas as pd

import torch
from torch import nn
import torch.nn.functional as F
from torchsummary import summary
import torch.backends.cudnn as cudnn
import torch.optim as optim


class Residual_block(nn.Module):
    def __init__(self, in_channels, out_channels, stride=1):
        super(Residual_block, self).__init__()
        
        self.residual = nn.Sequential(
            nn.Conv1d(in_channels=in_channels, out_channels=out_channels, kernel_size=1, stride=stride, padding=0, bias=False),
            nn.BatchNorm1d(out_channels)
        )

    def forward(self, x):
        x = self.residual(x)

        return x

class Inception_block(nn.Module):

  def __init__(self, in_channels, out_channels, stride=1, use_residual=True, use_bottleneck=True):
    super(Inception_block, self).__init__()
    
    self.use_bottleneck = use_bottleneck
    self.in_channels = in_channels
    self.out_channels = out_channels 
    self.bottleneck_size = 32


    if self.in_channels == 1:
        self.in_channels = int(in_channels)
    else:
        self.in_channels = int(self.bottleneck_size)
        self.bottleneck =  nn.Conv1d(in_channels, self.bottleneck_size, kernel_size=1, padding='same', bias=False)

    self.branch1 = nn.Conv1d(in_channels=self.in_channels, out_channels=self.out_channels, kernel_size=40,  stride=1, padding='same', bias=False)
    self.branch2 = nn.Conv1d(in_channels=self.in_channels, out_channels=self.out_channels, kernel_size=20,  stride=1, padding='same', bias=False)
    self.branch3 = nn.Conv1d(in_channels=self.in_channels, out_channels=self.out_channels, kernel_size=10,  stride=1, padding='same', bias=False)

    self.branch4 = nn.Sequential(
        nn.MaxPool1d(kernel_size=3, stride=1, padding=1),
        nn.Conv1d(in_channels=in_channels, out_channels=self.out_channels, kernel_size=1, stride=1, padding='same', bias=False),
    )

    self.bn = nn.BatchNorm1d(4*self.out_channels)
    self.relu = nn.ReLU()

  def forward(self, x):
    if self.use_bottleneck and int(x.shape[-2]) > 1:
        y = self.bottleneck(x)
    else:
        y = x        

    x = torch.cat([self.branch1(y), self.branch2(y), self.branch3(y), self.branch4(x)], 1)
    x = self.bn(x)
    x = self.relu(x)

    return x

class Classifier_INCEPTION(nn.Module):

    def __init__(self, input_shape, nb_classes, nb_filters=32, depth=6, residual=True):
        super(Classifier_INCEPTION, self).__init__()
        self.nb_classes = nb_classes
        self.nb_filters = nb_filters
        self.residual = residual
        self.depth = depth

        self.inception, self.shortcut = nn.ModuleList(), nn.ModuleList()
        
        for d in range(depth):
            self.inception.append(Inception_block(1 if d == 0 else nb_filters * 4, nb_filters))
            
            if self.residual and d % 3 == 2: 
                self.shortcut.append(Residual_block(1 if d == 2 else 4*nb_filters, 4*nb_filters))


        self.avgpool1 = nn.AdaptiveAvgPool1d(1)
        self.fc1 = nn.Linear(4*nb_filters, self.nb_classes)

        self.relu = nn.ReLU()

    def forward(self, x):
        input_res = x

        for d in range(self.depth):
            x = self.inception[d](x)
            if self.residual and d % 3 == 2:
                y = self.shortcut[d//3](input_res)
                x = self.relu(x + y)
                input_res = x

        x = self.avgpool1(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc1(x)

        return x




def fit(trainloader, valloader, input_shape, nb_classes):
    print('Hi from Inception!')
    recupereTeacherLossAccurayTest2 = []
    teacher = Classifier_INCEPTION(input_shape, nb_classes)

    print(teacher)

    use_cuda = torch.cuda.is_available()                
    if use_cuda:
        torch.cuda.set_device(0)
        teacher.cuda()
        cudnn.benchmark = True

    summary(teacher, ( input_shape[-2], input_shape[-1]))
    

    # Training
    def train_alone_model(net, epoch):

        print('\Teaining epoch: %d' % epoch)
        net.train()

        train_loss = 0
        correct = 0
        total = 0
        
        for batch_idx, (inputs, targets) in enumerate(trainloader):
            if use_cuda:
                inputs, targets = inputs.cuda(), targets.cuda()
            optimizer.zero_grad()

            outputs = net(inputs.float())
            loss = criterion_CE(outputs, targets)
            loss.backward() 
            optimizer.step()  

            train_loss += loss.item()

            _, predicted = torch.max(outputs.data, 1)
            total += targets.size(0)
            correct += predicted.eq(targets.data).cpu().sum().float().item()

            b_idx = batch_idx

        print('Training Loss: %.3f | Training Acc: %.3f%% (%d/%d)' % (train_loss / (b_idx + 1), 100. * correct / total, correct, total))

        return train_loss / (b_idx + 1)

    def test(net):
        net.eval()
        test_loss = 0
        correct = 0
        total = 0
        for batch_idx, (inputs, targets) in enumerate(valloader):
            if use_cuda:
                inputs, targets = inputs.cuda(), targets.cuda()
            outputs = net(inputs.float())
            loss = criterion_CE(outputs, targets)
            test_loss += loss.item()
            _, predicted = torch.max(outputs.data, 1)
            total += targets.size(0)
            correct += predicted.eq(targets.data).cpu().sum().float().item()
            b_idx = batch_idx

        print('Test Loss: %.3f | Test Acc: %.3f%% (%d/%d)' % (test_loss / (b_idx + 1), 100. * correct / total, correct, total))
        return test_loss / (b_idx + 1), correct / total

    final_loss = []
    learning_rates = []
    
    epochs = 15
    criterion_CE = nn.CrossEntropyLoss()
    optimizer = optim.Adam(teacher.parameters(), lr=0.001,)
    scheduler = optim.lr_scheduler.OneCycleLR(optimizer, 1e-3, epochs=epochs, steps_per_epoch=len(trainloader))

    use_cuda = torch.cuda.is_available()        
    best_model_wts = copy.deepcopy(teacher.state_dict())
    min_train_loss = np.inf

    start_time = time.time()    
    for epoch in range(epochs):
        train_loss = train_alone_model(teacher, epoch)
        test(teacher)
        if min_train_loss  > train_loss:
            min_train_loss = train_loss
            best_model_wts = copy.deepcopy(teacher.state_dict())
        scheduler.step()
    
        final_loss.append(train_loss)
        learning_rates.append(optimizer.param_groups[0]['lr'])
    
    output_directory = '/'
    torch.save(best_model_wts, output_directory +  'best_teacher_model.pt')


    # Save Logs
    duration = time.time() - start_time
    best_teacher = Classifier_INCEPTION(input_shape, nb_classes)
    best_teacher.load_state_dict(best_model_wts)
    best_teacher.cuda()
    
    print('Best Model Accuracy in below ')
    start_test_time = time.time()
    test(best_teacher)
    test_duration = time.time() - start_test_time

    print(test(best_teacher))

    df = pd.DataFrame(list(zip(final_loss, learning_rates)), columns =['loss', 'learning_rate'])
    index_best_model = df['loss'].idxmin()
    row_best_model = df.loc[index_best_model]
    df_best_model = pd.DataFrame(list(zip([row_best_model['loss']], [index_best_model+1])), columns =['best_model_train_loss', 'best_model_nb_epoch'])

    df.to_csv(output_directory + 'history.csv', index=False)
    df_best_model.to_csv(output_directory + 'df_best_model.csv', index=False)

    loss_, acc_ = test(best_teacher)
    df_metrics = pd.DataFrame(list(zip([min_train_loss], [acc_], [duration], [test_duration])), columns =['Loss', 'Accuracy', 'Duration', 'Test Duration'])
    df_metrics.to_csv(output_directory + 'df_metrics.csv', index=False)




import pandas as pd
import numpy as np
import sklearn
import os
from importlib import resources
from romaniya_menim.utils import transform_labels, read_all_datasets, prepare_data
import torch
from torch.utils.data import DataLoader, TensorDataset


def Classifier(datasets_dict, dataset_names, classifier_names):

    for dataset_name in dataset_names:
        for classifier_name in classifier_names:
            print('Classifier Type: ', classifier_name)   
            trainloader, valloader, input_shape, nb_classes = prepare_data(datasets_dict, dataset_name, classifier_name)
            
            if classifier_name == 'fcn':
                from classifiers import fcn
                print('Everything is ok and now FCN is calling')
                return fcn.create_FCN(trainloader, valloader, input_shape, nb_classes)

            elif classifier_name == 'cnn':
                from classifiers import cnn
                print('Everything is ok and now CNN is calling')
                return cnn.create_CNN(trainloader, valloader, input_shape, nb_classes)

            elif classifier_name == 'mlp':
                from classifiers import mlp
                print('Everything is ok and now MLP is calling')
                return mlp.create_MLP(trainloader, valloader, input_shape, nb_classes)

            elif classifier_name == 'resnet':
                from romaniya_menim.models import resnet
                print('Everything is ok and now RESNET is calling')
                return resnet.fit(trainloader, valloader, input_shape, nb_classes)

            elif classifier_name == 'inception':
                from romaniya_menim.models import inception
                print('Everything is ok and now INCEPTION is calling')
                return inception.fit(trainloader, valloader, input_shape, nb_classes)